# IssueTrackerProduct
#
# Peter Bengtsson <mail@peterbe.com>
#

__doc__="""IssueTrackerProduct is the easiest bug/issue tracker
system to use for Zope.
By Peter Bengtsson <mail@peterbe.com>

Credits:
Jo Meder for addhrefs() function"""
__version__='0.4.8a'

# python
import string
import os
import re
import sys
import random
import copy
from types import IntType, StringType, ListType

# Zope
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Globals import Persistent, InitializeClass, package_home, DTMLFile
from Products.PythonScripts.standard import html_quote, url_quote_plus
from Products.PythonScripts.PythonScript import PythonScript
from OFS import SimpleItem, ObjectManager, Folder
from DocumentTemplate import sequence
from AccessControl import ClassSecurityInfo
from Products.ZCatalog.CatalogAwareness import CatalogAware
from zLOG import LOG, ERROR, INFO, PROBLEM, WARNING
from DateTime import DateTime
# Product

import IssueTrackerNotifyables
import IssueTrackerUtils
from IssueTrackerUtils import cookIdAndTitle, ss, getNowdate,  \
                              fixDictofLists, getFromstring
from IssueTrackerConstants import *


manage_addIssueTrackerForm = PageTemplateFile('zpt/addIssueTrackerForm', globals())

def manage_addIssueTracker(dispatcher, id, title='', REQUEST=None):
    """ add IssueTracker instance via the web """
    dest = dispatcher.Destination()
    issuetracker = IssueTracker(id, title)
    dest._setObject(id, issuetracker)
    self = dest._getOb(id)
    
    
    self.DeployStandards()
    self.InitZCatalog()

    if REQUEST is not None:
        # whereto next?
        if REQUEST.has_key('addandedit'):
            REQUEST.RESPONSE.redirect(self.absolute_url()+'/manage_workspace')
        elif REQUEST.has_key('DestinationURL'):
            REQUEST.RESPONSE.redirect(REQUEST.DestinationURL+'/manage_workspace')
        else:
            REQUEST.RESPONSE.redirect(REQUEST.URL1+'/manage_workspace')
    
    

class IssueTracker(Folder.Folder, CatalogAware, Persistent,
                   IssueTrackerNotifyables.Notifyables
                   ):
    """ IssueTracker class """

    meta_type = ISSUETRACKER_METATYPE

    security=ClassSecurityInfo()

    manage_options = Folder.Folder.manage_options[:2] + \
        ({'label':'Properties', 'action':'manage_editIssueTrackerPropertiesForm'},
         {'label':'Management', 'action':'manage_ManagementForm'}) \
        + Folder.Folder.manage_options[3:]


    
    def __init__(self, id, title='',
                 sitemaster_name=DEFAULT_SITEMASTER_NAME,
                 sitemaster_email=DEFAULT_SITEMASTER_EMAIL):
        """ Init IssueTracker class """
        self.id = str(id)
        self.title = str(title)
        
        self.types              = DEFAULT_TYPES
        self.urgencies          = DEFAULT_URGENCIES
        self.sections_options   = DEFAULT_SECTIONS_OPTIONS
        self.defaultsections    = DEFAULT_SECTIONS
        self.when_ignore_word   = DEFAULT_WHEN_IGNORE_WORD
        self.display_date       = DEFAULT_DISPLAY_DATE
        self.always_notify      = DEFAULT_ALWAYS_NOTIFY
        self.sitemaster_name    = sitemaster_name
        self.sitemaster_email   = sitemaster_email
        self.odd_bgcolor        = DEFAULT_ODD_BGCOLOR
        self.even_bgcolor       = DEFAULT_EVEN_BGCOLOR
        self.default_type       = DEFAULT_TYPE
        self.default_urgency    = DEFAULT_URGENCY
	self.manager_roles 	= DEFAULT_MANAGER_ROLES
        self.default_batch_size = DEFAULT_DEFAULT_BATCH_SIZE
        self.issueprefix        = DEFAULT_ISSUEPREFIX
        self.no_fileattachments = DEFAULT_NO_FILEATTACHMENTS
        self.statuses           = DEFAULT_STATUSES
        self.display_formats    = DEFAULT_DISPLAY_FORMATS
        self.default_display_format = DEFAULT_DEFAULT_DISPLAY_FORMAT
        self.dispatch_on_submit = DEFAULT_DISPATCH_ON_SUBMIT
        self.randomid_length = DEFAULT_RANDOMID_LENGTH


    # forms
    manage_editIssueTrackerPropertiesForm = \
                    DTMLFile('dtml/editIssueTrackerPropertiesForm',\
                             globals(), native_properties=NATIVE_PROPERTIES)
    manage_ManagementForm = DTMLFile('dtml/ManagementForm', globals())

    # scripts

    def timeSince(self, firstdate, seconddate, afterword=None):
        """ wrap script """
        script = IssueTrackerUtils.timeSince
        return script(firstdate, seconddate, afterword)
    
    def ShowDescription(self, text, display_format=''):
        """ pass on to utilities module """
        script = IssueTrackerUtils.ShowDescription
        return script(text, display_format)

    def ShowFilesize(self, bytes):
        """ pass on to utilities module """
        return IssueTrackerUtils.ShowFilesize(bytes)

    ## This is "under construction"
    def standard_error_message(self, REQUEST, RESPONSE=None):
        """ custom error message """
        return getattr(self, 'standard_error_message.html')
                       

    def manage_editIssueTrackerProperties(self, REQUEST):
        """ save all IssueTracker related issues """
        hk = self.REQUEST.has_key
        get = self.REQUEST.get
        strings = ['title','when_ignore_word','display_date','sitemaster_name',
                   'sitemaster_email','odd_bgcolor','even_bgcolor',
                   'default_type','default_urgency','issueprefix',
                   'default_display_format']
        lists = ['types','urgencies','sections_options','defaultsections',
                 'statuses','display_formats','manager_roles',
                 'always_notify']
        ints = ['default_batch_size','randomid_length','no_fileattachments']
        booleans = ['dispatch_on_submit']
        for each in strings:
            if hk(each) and type(get(each))==StringType:
                exec 'self.%s = "%s"' % (each, get(each).strip())
        for each in ints:
            if hk(each):
                if type(get(each))==IntType:
                    exec "self.%s = %s" % (each, get(each))
                else:
                    LOG(self.meta_type, WARNING,
                        '%s not integer'%get(each), '')
        for each in lists:
            if hk(each) and type(get(each))==ListType:
                dummy = []
                for item in get(each):
                    if item.strip() != '':
                        dummy.append(item.strip())
                exec "self.%s = %s" % (each, dummy)
        for each in booleans:
            if hk(each) and get(each):
                exec 'self.%s = 1'%each
            else:
                exec 'self.%s = 0'%each
                
        # for the custom properties
        self.manage_editProperties(REQUEST)
        return self.manage_editIssueTrackerPropertiesForm(REQUEST,
                        manage_tabs_message='IssueTracker properties updated.')

    def DeployStandards(self, remove_oldstuff=0, DestinationURL=None,
                        initzcatalog=0):
        """ copy images and other documents into the instance unless they
            are already there
        """
        t={}
        
        if initzcatalog:
            self.InitZCatalog()
            
            
        # create folders
        root = self.getRoot()
        for f in ['notifyables', 'www']:
            if not hasattr(root, f):
                root.manage_addFolder(f)
                t[f]='Folder'

        osj = os.path.join
        standards_home = osj(package_home(globals()),'standards')
        t = self._deployImages(root, standards_home,
                               t=t, remove_oldstuff=remove_oldstuff)

        www_home = osj(standards_home,'www')
        t = self._deployImages(root.www, www_home,
                               t=t, remove_oldstuff=remove_oldstuff)

        # shortcut
        addPG = root.manage_addProduct['PageTemplates'].manage_addPageTemplate
        
        for filestr in os.listdir(standards_home):
            if filestr[-5:] == '.dtml':
                id, title = cookIdAndTitle(filestr.replace('.dtml',''))
                if hasattr(root, id) and remove_oldstuff:
                    root.manage_delObjects([id])
                
                if not hasattr(root, id):
                    file = DTMLFile('standards/%s'%filestr.replace('.dtml',''), globals()).read()
                    root.manage_addDTMLDocument(id, title, file=file)
                    t[id] ="DTML Document"
            elif filestr[-4:] == '.zpt':
                id, title = cookIdAndTitle(filestr.replace('.zpt',''))
                if hasattr(root, id) and remove_oldstuff:
                    root.manage_delObjects([id])
                    
                if not hasattr(root, id):
                    file = open(osj(standards_home,filestr)).read()
                    addPG(id, title=title, text=file)
                    t[id]="Page Template"                    
            elif filestr[-3:] == '.py':
                id, title = cookIdAndTitle(filestr.replace('.py',''))
                if hasattr(root, id) and remove_oldstuff:
                    root.manage_delObjects([id])
                    
                if not hasattr(root, id):
                    file = open(osj(standards_home, filestr)).read()
                    id = root._setObject(id, PythonScript(id))
                    root._getOb(id).write(file)
                    t[id]="Script (Python)"
        if DestinationURL:
            msg = ''
            for k,v in t.items():
                msg = "%s (%s) <br>%s"%(k,v,msg)
            method = IssueTrackerUtils.AddParam2URL
            url = method(DestinationURL,{'manage_tabs_message':\
                                "Standard objects deployed<br>%s"%msg})
            self.REQUEST.RESPONSE.redirect(url)
        else:
            return "Standard objects deployed\n%s"%t

    def _deployImages(self, destination, dir, extension='.gif', t={},
                      remove_oldstuff=0):
        """ do the actual deployment of images in a dir """
        osj = os.path.join
        for filestr in os.listdir(dir):
            if filestr.find(extension) > -1:
                # take the image
                id, title = cookIdAndTitle(filestr)
                if hasattr(destination, id) and remove_oldstuff:
                    destination.manage_delObjects([id])
                    
                if not hasattr(destination, id):
                    destination.manage_addImage(id, title=title, \
                          file=open(osj(dir, filestr),'rb').read())
                    t[id]="Image"
        return t

                
    def InitZCatalog(self):
        """ create a ZCatalog called 'Catalog' and change its properties
        accordingly """
        if not 'Catalog' in self.objectIds('ZCatalog'):
            self.manage_addProduct['ZCatalog'].manage_addZCatalog('Catalog','')
        zcatalog = self.Catalog
        indexes = zcatalog._catalog.indexes
        
        #shortcut
        pluginindexes = zcatalog.Indexes.manage_addProduct['PluginIndexes']
        
        add=('status','description','comment','fromname','email','url2issue')
        for adder in add:
            if not indexes.has_key(adder):
                pluginindexes.manage_addTextIndex(adder)


    def getRoot(self):
        """ Get the root instance object """
        if self.meta_type == ISSUETRACKER_METATYPE:
            return self
        else:
            for each in self.superValues(ISSUETRACKER_METATYPE):
                if each.id in self.REQUEST.URL1.split('/'):
                    return each

    def getRootURL(self):
        """ quick wrapper around getRoot() """
        return self.getRoot().absolute_url()

    def getCookiekey(self, which):
        """ return the cookiekey constants depending on key """
        which = which.lower().strip()
        if which=='name':
            return NAME_COOKIEKEY
        elif which=='email':
            return EMAIL_COOKIEKEY
        elif which=='display_format':
            return DISPLAY_FORMAT_COOKIEKEY
        
    def __before_publishing_traverse__(self, object, request):
        """ sort things out before publising object """
        self.get_environ()

    def get_environ(self):
        """ Populate REQUEST as appropriate """
        request = self.REQUEST
        stack = request['TraversalRequestNameStack']
        popped = []

        _special = 'REQUEST'
        # things to pop out
        queryitems = ({'key':'start', 'mkey':'start', 'type':'int'},
                      {'key':'sortorder', 'mkey':'sortorder', 'type':'string'},
                      {'key':'reverse', 'mkey':'reverse', 'type':'int'},
                      )

        splitter = '-'
        if stack:
            stack_copy = copy.deepcopy(stack)
            found_item = 1
            #while found_item and stack:
            for each in range(len(stack_copy)):
                found_item = 0
                stack_item = stack_copy[each]
                for each in queryitems:
                    key, value = each['key'], each.get('mkey')

                    if value is None and stack_item==key:
                        # this is a valueless item
                        found_item = 1
                        request.set(key, 1)
                    elif stack_item.startswith("%s%s"%(key,splitter)) \
                         and value==_special:
                        found_item = 1
                        first_key = stack_item.replace("%s%s"%(key,splitter),'')
                        
                        try:
                            key, value = first_key.split(splitter,1)
                            value = int(value)
                            request.set(key, value)
                        except ValueError:
                            try:
                                key, value = first_key.split(splitter,1)
                                request.set(key, value)
                            except:
                                pass
                            
                    elif stack_item.startswith("%s%s"%(key,splitter)):
                        found_item = 1
                        if each['type']=='int':
                            replace_what = "%s%s"%(key,splitter)
                            key = int(stack_item.replace(replace_what,''))
                        else:
                            replace_what = "%s%s"%(key,splitter)
                            key = stack_item.replace(replace_what,'')
                        
                        request.set(value, key)

                if found_item:
                    stack.remove(stack_item)
                    popped.append(stack_item)

            request.set('popped',popped)

    def aurl(self, url, params={}):
        """ modify the URL to include url-request-variables """
        request = self.REQUEST
        
        splitted = url.split('/')
        
        queryitems = ({'key':'start', 'mkey':'start'},
                      {'key':'sortorder', 'mkey':'sortorder'},
                      {'key':'reverse', 'mkey':'reverse'},
                      )
        
        splitter = '-'
        # Use old things
        
        keys_applied = []
        for key, value in params.items():
            keys_applied.append(key)
            if value is not None:
                splitted.append("%s%s%s"%(key, splitter, value))

        for each in queryitems:
            key, mkey = each['key'], each.get('mkey')
            if mkey is not None:
                if key not in keys_applied and request.has_key(mkey):
                    splitted.append("%s%s%s"%(key, splitter, request[mkey]))
        
            
        
        return '/'.join(splitted)


    def _fillRequest(self, dict):
        """ add properties as empty if not existing """
        propertiesNulls	= ['title','fromname','email','description',
                           'url2issue','urgency','type','display_format']
        propertiesZero 	= ['confidential','hide_me']
        propertiesEmpty	= ['sections']
        
        nullified = []
        for property in propertiesNulls:
            if not dict.has_key(property):
                dict[property] =''
                
        for property in propertiesZero:
            if not dict.has_key(property):
                dict[property] = 0
                
        for property in propertiesEmpty:
            if not dict.has_key(property):
                dict[property]=[]
        
        # Also, for all list variables, make sure that only
        # one item doesn't make it a string
        for property in propertiesEmpty:
            if type(dict[property]) == StringType:
                dict[property] = [dict[property]]
                

        # special one
        try:
            if not dict['url2issue'].startswith('/') and \
               not dict['url2issue'].startswith('http') and \
               not dict['url2issue'].strip()=='':
                dict['url2issue'] = 'http://%s'%dict['url2issue']
        except:pass # no big deal

        return dict

    
    def getFileIconpath(self, filename):
        """ Try to find a suitable file icon """
        default = '/misc_/OFSP/File_icon.gif'
        extension = filename.lower()[filename.rfind('.')+1:]
        if ICON_ASSOCIATIONS.has_key(extension):
            return '%s/%s'%(ICON_LOCATION,ICON_ASSOCIATIONS[extension])
        else:
            return default

    def getFileattachmentInput(self, index):
        """ return either a file input field or a keep option """
        request = self.REQUEST

        
        input_field = """<input size="%s" 
        name="fileattachment:list"  type="file">"""
        icon_html = """<img hspace="2" src="%s" alt="File"
        title="File size %s" border="0" />"""
        if request.has_key('Tempfolder_fileattachments'):
            upload_folder = request['Tempfolder_fileattachments']
            # Maybe the actual folder doesn't exist any more
            if not hasattr(self.temp_folder, upload_folder):
                return input_field%'40'
            files = self.temp_folder[upload_folder].objectValues('File')
            try:
                file = files[index]
                file_src = self.getFileIconpath(file.getId())
                file_size = self.ShowFilesize(file.getSize())
                icon = icon_html%(file_src, file_size)
                
                return '<a href="%s" title="File size %s">%s%s</a>'%\
                       (file.absolute_url(), file_size, icon, file.getId())                
            except:
                return input_field%'40'
        else:
            return input_field%'40'
    
    def UploadTempFiles(self):
        """ Attempt to upload fileattachments to temp_folder and stick
        some information in the REQUEST """
        request = self.REQUEST
        temp_folder_id = None

        if request.has_key('Tempfolder_fileattachments'):
            temp_folder_id = request.get('Tempfolder_fileattachments')
        
        if request.has_key('fileattachment'):
            fileattachment = request.fileattachment
            # fileattachment is a list, deal with each item
            for file in fileattachment:
                if self._isFile(file):
                    if temp_folder_id is None:
                        temp_folder_id = self._generateTempFolder()

                    temp_folder = self.temp_folder[temp_folder_id]
                    filename = getattr(file, 'filename')
                    id=filename[max(string.rfind(filename, '/'),
                                    string.rfind(filename, '\\'),
                                    string.rfind(filename, ':'),
                                    )+1:]
                    if id.startswith('_'):
                        id=id[1:]
                    temp_folder.manage_addFile(id=id, file=file)
            # This tests whether any files were uploaded
            if temp_folder_id is not None:
                request.set('Tempfolder_fileattachments', temp_folder_id)

        return "anything"

    def _isFile(self, file):
        """ Check if Publisher file is a real file """
        try:
            if hasattr(file, 'filename'):
                return getattr(file, 'filename').strip() != ''
            else:
                return 0
            
        except:
            return 0
        
    def _generateTempFolder(self):
        """ Create a folder in temp_folder with randomish id and return
        its id """
        root = self.temp_folder
        timestamp = str(int(self.ZopeTime()))
        randstr = self.getRandomString(length=3, numbersonly=1)
        rand_id_start = "uploadtmp-it-%s"%timestamp
        rand_id = "%s-%s"%(rand_id_start, randstr)
        while hasattr(root, rand_id):
            new_rand_str = self.getRandomString(length=3, numbersonly=1)
            rand_id = "%s-%s"%(rand_id_start, new_rand_str)

        root.manage_addFolder(rand_id)
        return rand_id

    def getFileattachmentContainer(self):
        """ if 'Tempfolder_fileattachments' is set in REQUEST return folder
        object, otherwise return self. """
        request = self.REQUEST

        if request.has_key('Tempfolder_fileattachments'):
            return self.temp_folder[request['Tempfolder_fileattachments']]
        else:
            return self
        
    def SubmitIssue(self):
        """ This is the method to create an Issue Tracker Issue. It
        relies only on the REQUEST object.
        1) Check data
        2) Try to create issue
            2a) If success, RESPONSE.redirect to issue plus Thank you message
            2b) If failure, print failed data and urge to submit again
        """

        request = self.REQUEST

        failures = {}
        has_cookie = self.has_cookie
        get_cookie = self.get_cookie
        set_cookie = self.set_cookie

        # check for existance, by adding all missing properties
        # as empty instead of non-existant.
        request.set('title',request['subject'])
        request = self._fillRequest(request)
        
        #
        # Tune the data a bit
        #

        # strip whitespace
        for property in ['title','fromname','email',
                         'url2issue','display_format']:
            request[property] = request[property].strip()
        # Special treatment needed in case STX is used upon display
        request['description'] = request['description'].strip()+' '

        email_cookiekey = self.getCookiekey('email')
        name_cookiekey = self.getCookiekey('name')
        display_format_cookiekey = self.getCookiekey('display_format')
        # use cookie if not else specified
        if request.email=='' and has_cookie(email_cookiekey):
            request['email'] = get_cookie(email_cookiekey)
        if request.fromname=='' and has_cookie(name_cookiekey):
            request['fromname'] = get_cookie(name_cookiekey)
        if request.display_format=='' and has_cookie(display_format_cookiekey):
           request['display_format'] = get_cookie(display_format_cookiekey)

        # append the default sections if not specified
        if len(request.sections) == 0:
            request.sections = self.defaultsections

        #
        # Check data
        #

        if request.subject == '':
            failures[FAILURE_SUBJECT] = FAILURE_SUBJECT_DESC

        valid_emailaddress = 1
        # to prevent problems with sending mail
        if not IssueTrackerUtils.ValidEmailAddress(request.email):
            valid_emailaddress = 0

        if len(failures) >0:
            return self.ExplainSubmitErrors(failures)

        #
        # Let's submit the issue!
        #

        prefix = self.issueprefix
        genid = self.generateID(self.randomid_length, prefix)

        # if these are valid, save them
        if request.fromname <> '':
            set_cookie(self.getCookiekey('name'), request.fromname)
        if request.display_format in self.display_formats:
            set_cookie(self.getCookiekey('display_format'),request.display_format)
        if valid_emailaddress:
            set_cookie(self.getCookiekey('email'), request.email)


        # filter out empty item from sections
        sections_newlist=[]
        for section in request.sections:
            if section!='':
                sections_newlist.append(section)

        # add all the properties
        title           = request.title
        fromname	= request.fromname
        email    	= request.email
        url2issue 	= request.url2issue
        type     	= request.type
        urgency 	= request.urgency
        description 	= request.description
        display_format  = request.display_format
        confidential 	= int(request.confidential)
        hide_me 	= int(request.hide_me)
        issuedate 	= getNowdate()
        status  	= request.get('status','open')
        sections        = sections_newlist


        # Do the actual object adding
        issue = self._createIssueObject(genid, request.title, status, type, urgency,
                                  sections, fromname, email, url2issue,
                                  confidential, hide_me, description,
                                  display_format, issuedate)
        

        # catalog it
        issue.index_object()

        # remember it
        self.RememberAddedIssue(genid)
        

        # Also upload the fileattachments
        if request.has_key('Tempfolder_fileattachments'):
            files_copied = []
            upload_folder_id = request.get('Tempfolder_fileattachments')
            upload_folder = self.temp_folder[upload_folder_id]
            for file in upload_folder.objectValues('File'):
                issue.manage_clone(file, file.getId())
                files_copied.append(file.getId())
            upload_folder.manage_delObjects(files_copied)
            self.temp_folder.manage_delObjects([upload_folder_id])

        if request.has_key('fileattachment'):
            fileattachment = request.fileattachment
            for file in fileattachment:
                if self._isFile(file):
                    filename = getattr(file, 'filename')
                    id=filename[max(string.rfind(filename, '/'),
                                    string.rfind(filename, '\\'),
                                    string.rfind(filename, ':'),
                                    )+1:]
                    if id.startswith('_'):
                        id=id[1:]
                    issue.manage_addFile(id=id, file=file)

        always_emailstring = ', '.join(self.getAlwaysNotify())
        self.SendIssue2Friends_script(genid, always_emailstring)

        # Where to next?
        redirect_url = '%s?NewIssue=Submitted'%(issue.absolute_url())
        redirect_url = '%s&NotYetRecent=%s'%(redirect_url, genid)
        request.RESPONSE.redirect(redirect_url)


    def _createIssueObject(self, id, title, status, type, urgency, sections,
                     fromname, email, url2issue, confidential, hide_me,
                     description, display_format, issuedate, index=0):
        """ crudely create issue object. No checking """
        issue = IssueTrackerIssue(id, title, status, type, urgency,
                                  sections, fromname, email, url2issue,
                                  confidential, hide_me, description,
                                  display_format, issuedate)

        # not here
        self._setObject(id, issue)
        issue = getattr(self, id)

        if index:
            # catalog it
            issue.index_object()

        return issue

    def generateID(self, length, prefix='',
                    meta_type=ISSUE_METATYPE,increment=None):
        """ generate IDs for different objects """
        if increment is None:
            idnr = len(self.objectIds(meta_type))+1
            increment = idnr + 1
        else:
            idnr = increment
            increment = increment +1
        id='%s%s'%(prefix, string.zfill(str(idnr), length))
        base= getattr(self,'aq_base',self)
        if hasattr(base, id):
            # ah! Id already exists, try again
            return self.generateID(length, prefix, meta_type, 
                                   increment=increment)
        else:
            return id


    def About(self):
        """ Show some info about the product """
        osj = os.path.join
        About_content = open(osj(package_home(globals()), 'About.html')).read()
        f = open(osj(package_home(globals()), 'CHANGES.txt'), "r")
        changelog = f.read()
        f.close()
        changelog = "<PRE>%s</PRE>"%self.ShowDescription(changelog, 'structured_text')
        About_content = About_content.replace('*CHANGELOG*',changelog)
        About_content = About_content.replace('*VERSION*',__version__)
        return About_content

    # Cookies!

    def get_cookie(self, name, default=None):
        """ return RESPONSE cookie """
        return self.REQUEST.cookies.get(name,default)
    
    def set_cookie(self, name, value):
        """ set RESPONSE cookie """
        later = str(IssueTrackerUtils.getLaterDate(365))
        self.REQUEST.RESPONSE.setCookie(name,value, expires=later, path='/')
        
    def has_cookie(self, name):
        """ return cookie presence """
        return self.REQUEST.cookies.has_key(name)


    # Misc

    def url_quote_plus(self, s):
        """ wrap the url_quote_plus script from standards """
        return url_quote_plus(s)
    
    def UpdateCatalog(self):
        """ Re-find items in the Catalog """
        request = self.REQUEST

        if request.has_key('URL2'):
            URL2 = request.URL2
        else:
            URL2 = request.URL1

        method = self.Catalog.manage_catalogFoundItems
        method(request, request.RESPONSE, URL2, request.URL1,
               obj_metatypes=[ISSUE_METATYPE])
        
    def issueURLbyID(self, issueID):
        """ Return absolute_url of an issue from its id """
        return getattr(self.getRoot(),issueID).absolute_url()

    def getAllNotifications(self):
        """ Go through all issues and find all notification objects """
        all = []
        for issue in self.getRoot().objectValues(ISSUE_METATYPE):
            all = all+issue.objectValues('Issue Notification')
        return all
    
    def LineIndent(self, text, indent):
        """ wrap script """
        return IssueTrackerUtils.LineIndent(text, indent)
    
    def preParseEmailString(self, email_string):
        """ wrapper around utils """
        parsemethod = IssueTrackerUtils.preParseEmailString
        all_notifyables = self.getNotifyables()
        if all_notifyables:
            names2emails = {}
            for item in all_notifyables:
                names2emails[item.getName()] = item.getEmail()
            result = parsemethod(email_string, names2emails=names2emails)
            
        else:
            result = parsemethod(email_string)
        return result
       

    def getManagerRoles(self):
	""" Return the roles that makes an IssueTracker Manager """
	if hasattr(self, 'manager_roles'):
	    return self.manager_roles
	else:
            return ['Manager'] # backward compatability

    def hasManagerRole(self):
        """
            This method determines if the current user
            is allowed to do stuff that only the Zope manager is
            supposed to be able to do.
            Feel free to edit appropriatly to what suits you.
        """
	user_roles = self.REQUEST.AUTHENTICATED_USER.getRoles() 
	for role in self.getManagerRoles():
	    if role in user_roles:
		return 1
	# still here!
	return 0

    def getRandomString(self, length=5, loweronly=0, numbersonly=0):
        """ return a completely random piece of string """
        script = IssueTrackerUtils.getRandomString
        return script(length, loweronly, numbersonly)

    
    def thisInURL(self, page, homepage=0):
        """ To find if a certain objectid is in the URL """
        URL = self.ActionURL(self.REQUEST.URL)
        rootURL = self.getRootURL()
        if homepage and URL==rootURL:
            return true
        else:
            expected = rootURL +'/'+ page
            if URL.lower() == expected.lower():
                return true
            else:
                return false

    def ActionURL(self, url=None):
        """
        If URL is http://host/index_html
        I prefer to display it http://host
        Just a little Look&Feel thing
        """
        if url is None:
            url = self.REQUEST.URL

        URLsplitted = string.split(url,'/')
        if URLsplitted[-1] == 'index_html':
            return string.join(URLsplitted[:-1],'/')

        return url

    def ExplainSubmitErrors(self, dict):
        """ Return HTMLed statements about errors """
        li = []
        for key,value in dict.items():
            li.append('<strong>%s</strong> - %s'%(key, value))
        return li

    def ShowFilter(self, filtername, sequence=[]):
        """ Check whether to show filter or not """
        request = self.REQUEST
	trackerroot = self.getRoot()
        if request.has_key(filtername):
            return request[filtername]
        elif not request.has_key('%s_filteroptions'%trackerroot.id):
            return []

        return []

    def hasStoredFilter(self):
        """ Check if filter is applied """
	trackerroot = self.getRoot()
        return self.REQUEST.has_key('%s_filteroptions'%trackerroot.id)

    def ShowOnlyMe(self):
        """ enable listing with set name and email """
        request = self.REQUEST
        redirect = request.RESPONSE.redirect

        filters={}
        for each in ['fromname','email']:
            if request.has_key(each) and request[each] != '':
                filters[each]=request[each]

        if filters:
            querystring = []
            for k,v in filters.items():
                querystring.append('%s=%s'%(k,v))
            # make list into string
            querystring = '?'+string.join(querystring,'&')
        else:
            querystring=''

        redirect(self.getRootURL()+'/ListIssues'+querystring,lock=1)

    def ShowNameEmail(self, fromname, email, hideme=None, highlight=1):
        """ Show name and email depending on certain criterias """
        out = ''
        if highlight:
            fromname = self.HighlightQ(html_quote(fromname))
            email    = html_quote(email)
            show_email = self.HighlightQ(email)
            
        else:
            fromname = html_quote(fromname)
            email    = html_quote(email)
            show_email = email

        if fromname =='' and email=='':
            name_email = NONAME_NOEMAIL
        elif fromname =='': 
            name_email ='&lt;<a href="mailto:%s">%s</a>&gt;'%(email,show_email)
        elif email =='':
            name_email = fromname
        else:
            name_email = '%s &lt;<a href="mailto:%s">%s</a>&gt;'%(fromname, email, show_email)
         

        if hideme is not None and hideme:
            out += NAME_EMAIL_HIDDEN + ", "
            if self.hasManagerRole():
                out += name_email
        else:
            out += name_email

        return out

    def getIssueURL(self, issue, no_q=0):
        """ Return an issues URL but maybe add a thing to it """
        url = issue.absolute_url()
        request = self.REQUEST
        if no_q or not (request.has_key('q') and request['q']!=''):
            return url
        else:
            return "%s?q=%s"%(url, url_quote_plus(request['q']))

    def setFollowupCursor(self):
        """ Check if it's time to add a comment, and if that's the case
        prepare something in the REQUEST so that StandardLook can pick
        up on that. """
        if self.has_key_special('IssueAction', 1):
            textarea_name = "document.form_followup['comment_description:text']"
            self.REQUEST.set('FirstFocus', textarea_name)
            
    def setAddIssueCursor(self):
        """ If it's addIssue or quick, set the focus """
        textarea_name = "document.add_issue.subject"
        hk = self.REQUEST.has_key
        if not hk('previewissue'):
            self.REQUEST.set('FirstFocus', textarea_name)

    def getFirstFocusJavascript(self):
        """ create a script tag if 'FirstFocus' """
        request = self.REQUEST
        s="function setFirstFocus() {"
        if request.has_key('FirstFocus'):
            s=s+ request.get('FirstFocus')+'.focus();'
        s=s+'}'
        return s
        
    def has_key_special(self, name, shorten=0):
        """
        Normally you would do REQUEST.has_key('IssueAction')
        but if an imagebutton is used you'll find that you have
        REQUEST['IssueAction.y'] and REQUEST['IssueAction.x']
        But the result should be the same.
        """
        request = self.REQUEST
        if request.has_key(name):
            return 1
        elif request.has_key('%s.y'%name) and request.has_key('%s.x'%name):
            return 1
        elif shorten:
            for key in request.keys():
                if key[:len(name)]==name:
                    return 1
            return 0
        else:
            return 0

    def get_special_key(self, name):
        """
        Normally you would do REQUEST.has_key('IssueAction')
        but if an imagebutton is used you'll find that you have
        REQUEST['IssueAction.y'] and REQUEST['IssueAction.x']
        But the result should be the same.
        """
        try:
            return self.REQUEST[name]
        except KeyError:
            try:
                return self.REQUEST['%s.x'%name]
            except:
                raise KeyError, name

    def getSavedUser(self, name_email='email'):
        """
        Return the name or email from request,
        if not found, return from the session object,
        else return ""
        """
        request = self.REQUEST

        if name_email =='email':
            s='email'
            cookie=self.getCookiekey('email')
        else:
            s='fromname'
            cookie=self.getCookiekey('name')

        # now we know what we're looking for

        if request.has_key(s):
            return request[s]
        elif self.has_cookie(cookie):
            return self.get_cookie(cookie)

        return ""

    def getPreviewTitle(self, oldstatus, action):
        """ Get what the title of thread will be (via the web only) """
        request = self.REQUEST
        action = ss(action)
        addfollowup = 0
        past_tense=None
        if action == 'open':
            past_tense='open'
            addfollowup = 1
        if action == 'take':
            past_tense='taken'
            addfollowup = 1
        if action == 'put on hold':
            past_tense='on hold'
            addfollowup = 1
        if action == 'reject':
            past_tense='rejected'
            addfollowup = 1
        if action == 'complete':
            past_tense='completed'
            addfollowup = 1
        if action == 'delete':
            DeleteIssue(self, self.id)
            addfollowup = 0
            redirect_url = request.URL2
        if action == 'addfollowup':
            addfollowup = 1
            dummychange = 1
            req_manager = 0
        if action == 'addfollowup':
            gentitle = 'Added Issue followup'
        else:
            gentitle = 'Changed status from %s to %s'%\
                       (oldstatus.capitalize(), past_tense.capitalize())

        return gentitle


    def ManagerOptionsExtend(self):
        """ Determine which forms to show to Managers """
        request = self.REQUEST

        has_key_special = self.has_key_special
        get_special_key = self.get_special_key

        do_preview = 0
        if has_key_special('IssueAction') and has_key_special('previewissue'):
            do_preview = 1
            
        form = ''
        if has_key_special('IssueAction') and not do_preview:
            action = get_special_key('IssueAction')
            return getattr(self,request['issueID']).ModifyIssue(action=action)
        elif has_key_special('IssueAction_AddFollowup'):
            form = self.form_followup()
        elif has_key_special('IssueAction_Open'):
            form = self.form_open()
        elif has_key_special('IssueAction_Take'):
            form = self.form_take()
        elif has_key_special('IssueAction_Onhold'):
            form = self.form_onhold()
        elif has_key_special('IssueAction_Rejected'):
            form = self.form_reject()
        elif has_key_special('IssueAction_Complete'):
            form = self.form_complete()
        elif has_key_special('IssueAction_Delete'):
            form = self.form_delete()
        else:
            form = self.manager_options()

        return form

    def AnonymousOptionsExtend(self):
        """ Determine which forms to show to Anonymous """
        request = self.REQUEST

        has_key_special = self.has_key_special
        get_special_key = self.get_special_key

        do_preview = 0
        if has_key_special('IssueAction') and has_key_special('previewissue'):
            do_preview = 1

        form = ''
        if has_key_special('IssueAction') and not do_preview:
            action = get_special_key('IssueAction')
            return getattr(self,request['issueID']).ModifyIssue(action=action)
        elif has_key_special('IssueAction_AddFollowup'):
            form = self.form_followup()
        else:
            form = self.anonymous_options()

        return form


    def CountStatuses(self, since=None):
        """ Return how many Issues there are under each status since
            a certain time
        """
        #
        # NEEDS WORK!!
        #
        
        #return {}
        request = self.REQUEST

        # check that since isn't a string
        if type(since)==StringType:
            since = DateTime(since)
        elif request.has_key('count_status_since'):
            since = DateTime()-int(request['count_status_since'])

        res={}
        tres=[]
        for issue in self.objectValues(ISSUE_METATYPE):
            if since is None or issue.bobobase_modification_time() >= since:
                status = issue.status.lower()
                if res.has_key(status):
                    res[status]=res[status]+1
                else:
                    res[status] = 1

        # Lastly we want to organize res by self.statuses order
        for status in self.statuses:
            status = status.lower()
            if res.has_key(status):
                tres.append({status:res[status]})
        return tres

    def BatchedQueryString(self, batchdict):
        """
        return QUERY_STRING but make sure stuff in
        the batchdict isn't duplicated.
        """
        request = self.REQUEST
        batchdict = self._Zero2None(batchdict)
        url = self.aurl(self.ActionURL(), batchdict)
        url = self._addQuerystring(url)
        return url
    

    def _Zero2None(self, dict):
        """ Replace all occurances of 0 (as tested int) to None """
        n_dict={}
        for key, value in dict.items():
            try:
                if int(value)==0:
                    n_dict[key]=None
                else:
                    n_dict[key]=value
            except:
                n_dict[key]=value
        return n_dict
    
    def DeleteIssue(self):
        """ Delete an Issue from the IssueTracker instance """
        request = self.REQUEST

        if request.has_key('issueID') and self.hasManagerRole():
            self.manage_delObjects(request['issueID'])

            # delete all notifications about this Issue
            del_notify_ids = []
            for notifyobject in self.objectValues('Issue Notification'):
                if notifyobject.issueID == request['issueID']:
                    del_notify_ids.append(notifyobject.id)
          
            self.manage_delObjects(del_notify_ids)

            request.RESPONSE.redirect(request.URL1+'/ListIssues')
        else:
            raise "IssueTrackerError", "The issueID could not be found in the REQUEST"

    def standard_html_header(context):
        """ to make it possible to use DTML objects here """
        breakword = '<!--METALbody-->'
        page = context.StandardLook()
        return page[:page.find(breakword)]

    def standard_html_footer(context):
        """ to make it possible to use DTML objects here """
        breakword = '<!--METALbody-->'
        page = context.StandardLook()
        return page[page.find(breakword)+len(breakword)+1:]

    def ManagerLink(context):
        """ For the little hyperlink where you can login with """
        link = context.login.absolute_url()
        came_from = context.ActionURL()
        rurl=str(int(getNowdate()))[0:4]
        return "%s?came_from=%s&r.url=%s"%(link, came_from, rurl)

    def getSavedTextFormat(self, no_default=0):
        """
        This method returns what display_format value the user has. 
        If none found, then the default one is returned.
        """

        if no_default:
            default = ""
        else:
            default = self.default_display_format

        s=None
        cookiekey = self.getCookiekey('display_format')

        if self.has_cookie(cookiekey):
            s = self.get_cookie(cookiekey)

        if s not in self.display_formats:
            s = None

        if s is None:
            return default
        else:
            return s

    def getNoFileattachments(self):
        """ return no_fileattachments or default """
        if hasattr(self, 'no_fileattachments'):
            return self.no_fileattachments
        else:
            return DEFAULT_NO_FILEATTACHMENTS

    def getAlwaysNotify(self):
        """ return always_notify or default """
        if hasattr(self, 'always_notify'):
            return self.always_notify
        else:
            return DEFAULT_ALWAYS_NOTIFY

    def saveEmailstring(self, to):
	""" Save to string as a cookie """
	key = EMAILSTRING_COOKIEKEY
	key = self.defineInstanceCookieKey(key)
	self.set_cookie(key, to)

    def getSavedEmailstring(self):
	""" Return cookie translated or nothing """
	key = EMAILSTRING_COOKIEKEY
	key = self.defineInstanceCookieKey(key)
	
	if self.REQUEST.cookies.has_key(key):
	    to = self.REQUEST.cookies[key]
	    for item in self.getNotifyables():
                to = to.replace(item.getEmail(), item.getName())
	    return to
	else:
	    return None

    def sendEmail(self, body, to, fr, subject):
        """ Attempt to send emails un protectedly """
        if hasattr(self, 'MailHost'):
            mailhost = self.MailHost
        else:
            all_mailhosts = self.superValues('Mail Host')
            if all_mailhosts:
                mailhost = all_mailhosts[0] # first one
            else:
                raise "AttributeError", "MailHost not found"
        try:
            mailhost.send(body, to, fr, subject)
        except:
            type, val, tb = sys.exc_info()
            LOG(self.__class__.__name__, WARNING,
                'Could not send email to %s'%to, '')
            raise type, val
            
    


    # List issues
    #

    def ListIssuesFiltered(self):
        """ wrapper around _ListIssuesFiltered() that prepares a search
        if REQUEST holds 'q'
        """
        q=None
        catalog = self.Catalog

        root = self.getRoot()
        root_issues = root.objectValues(ISSUE_METATYPE)
        if self.REQUEST.has_key('q') and self.REQUEST['q'].strip() != '':
            q=self.REQUEST['q'].strip()
            transtab = string.maketrans('/ ','_ ')
            q=string.translate(q, transtab, '?&!;()<=>*#[]{}')
##            q=q.replace('%','*') # allow both wildcards # needs thought
        seq = []
        if q is not None:
            catalogs = catalog({'title':'*%s*'%q}) + \
                       catalog({'description':q}) + \
                       catalog({'fromname':q, 'email':q}) 
            if len(catalogs)==0:
                # nothing found, maybe user typed in an id
                catalogs = catalog({'id':q})
                if len(catalogs)==1 and catalogs[0].getObject() in root_issues:
                    response = self.REQUEST.RESPONSE
                    url = catalogs[0].getObject().absolute_url()
                    response.redirect(url)
                
        else:
            catalogs = catalog()

        for brain in catalogs:
            object = brain.getObject()
            if object not in seq and object in root_issues:
                seq.append(object)
        if q is not None:
            self.RememberSearchTerm(q, len(seq))
        return self._ListIssuesFiltered(seq)
        

    def _ListIssuesFiltered(self, objsequence):
        """ Basically filter a sequence of issue objects """
        rlist = []
        request = self.REQUEST
        request.set('TotalNoIssues', len(objsequence))

        set_cookie = self.set_cookie

        try:
            # if there and integer'able
            reverse=int(request.reverse)
        except:
            reverse=0

        trackerroot = self.getRoot()

        #TrackerFilteroptionsCookiename
        tfc = "%s_filteroptions"%trackerroot.id
        
        # remember filteroptions
        if request.has_key('filteroptions'):
            tracker_filteroptions={}
            for each in ['ignoretypes','ignoresections','ignorestatuses',
                         'ignoreurgencies','ignorefromname','ignoreemail',
                         'showfromname','showemail']:
                try:
                    tracker_filteroptions[each]=request[each]
                except:pass
            set_cookie(tfc,tracker_filteroptions)
            
       
        #
        # prepare a master filter
        #
        
        filter ={}
        
        # statuses
        if request.has_key('ignorestatuses'):
            filter['status'] = request.ignorestatuses
        elif request.has_key('status'):
            filter['status'] = self._pop_from_list(request['status'],
                                                   self.statuses)
        else:
            filter['status'] = []
            
        # sections
        if request.has_key('ignoresections'):
            filter['sections'] = request.ignoresections
        elif request.has_key('section'):
            filter['sections'] = [request.section.strip()]
        else:
            filter['sections'] = []

        # urgencies
        if request.has_key('ignoreurgencies'):
            filter['urgency'] = request.ignoreurgencies
        elif request.has_key('urgency'):
            filter['urgency'] = [request.urgency.strip()]
        else:
            filter['urgency'] = []
            
        # types
        if request.has_key('ignoretypes'):
            filter['type'] = request.ignoretypes
        elif request.has_key('type'):
            filter['type'] = [request.type.strip()]
        else:
            filter['type'] = []
            
        ## fromname
        # Either a name to ignore or show only.
        # It does not make sense to have both showfromname and ignorefromname
        if request.has_key('showfromname') and \
           request.showfromname <> self.when_ignore_word:
                filter['showfromname'] = request.showfromname.strip()
        elif request.has_key('ignorefromname') and \
             request.ignorefromname <> self.when_ignore_word:
            filter['ignorefromname'] = request.fromname.strip()

        # email
        if request.has_key('showemail') and \
           request.showemail <> self.when_ignore_word:
                filter['showemail'] = string.strip(request.showemail)
        elif request.has_key('ignoreemail') and \
           request.ignoreemail <> self.when_ignore_word:
            filter['ignoreemail'] = request.email


        # prepare for caseindependence
        filter = fixDictofLists(filter)
            
        # all Issue objects
        issues = objsequence 

        #
        # Before we can determine what to sort on
        # we need to check the name of the properties
        
        # the default
        sortorder = ('bobobase_modification_time',)
        skipsort = 0
        if request.has_key('sortorder') and request['sortorder'] != 'date':
            sortorder = (request['sortorder'],)
            if request['sortorder'] == 'from':
                sortorder = ('fromname',)
            if request['sortorder'] == 'creationdate':
                sortorder = ('issuedate',)
            if request['sortorder'] == 'changedate':
                sortorder = ('bobobase_modification_time',)

            if request['sortorder'] == 'status':
                #
                # Very special!
                # self.statuses is an "human ordered" list like 'open','taken',...'completed','forgotten'
                statuses = {}
                # create a dictionary of list divided in groups by their status
                # like this: {'open':['foo','bar'], ... 'forgotten':['this','that']}
                for issue in issues:
                    if statuses.has_key(issue.status):
                        statuses[issue.status].append(issue)
                    else:
                        statuses[issue.status] = [issue]
                # recreate the containers
                issues = []
                # remembers that self.statuses is ordered
                #
                # self.statuses is the property
                # statuses is the dictionary
                #
                for status in self.statuses:
                    # perhaps there are no 'forgotten' issues even
                    # though there is 'forgotten' in self.statuses
                    if statuses.has_key(status):
                        # order all issues with this status by bobobase_modification_time
                        thisstatus_containers = sequence.sort(statuses[status],(('bobobase_modification_time',),))
                        # reverse bobobase_modification_time sort naturally
                        thisstatus_containers.reverse()
                        issues = issues + thisstatus_containers
                        # skip the default sort later
                        skipsort = 1
                        
            if request['sortorder'] == 'urgency':
                #
                # Very special too
                # read the comments on "if request['sortorder'] == 'status':"
                # above.
                urgencies = {}
                for issue in issues:
                    if urgencies.has_key(issue.urgency):
                        urgencies[issue.urgency].append(issue)
                    else:
                        urgencies[issue.urgency] = [issue]
                issues = []
                for urgency in self.urgencies:#+():
                    if urgencies.has_key(urgency):
                        thisurgency_containers = sequence.sort(urgencies[urgency], (('bobobase_modification_time',),))
                        thisurgency_containers.reverse()
                        issues = issues + thisurgency_containers
                        skipsort = 1
                
        else:
            # if no sortorder, it uses the bobobase_modification_time
            # and this has to be reversed so that the latest is listed first
            reverse =1

        # do the basic sort.
        # 'sortorder' is a tuple
        if not skipsort:
            issues = sequence.sort(issues,(sortorder,))

            
        if reverse:
            issues.reverse()

        # containers is not order, just filter it
        for issue in issues:
            # if confidential and (if) Manager show, if not skip
            if not issue.IsConfidential() or self.hasManagerRole():
                if filter == {}:
                    rlist.append(issue)
                else:
                    passsections = 1
                    for section in issue.sections:
                        if ss(section) in filter['sections']:
                            passsections = 0
                            break
                    # pass the filter lists

                    if ss(issue.status) not in filter['status'] \
                       and ss(issue.urgency) not in filter['urgency'] \
                       and ss(issue.type) not in filter['type'] \
                       and passsections and \
                       self._filter_fromnameemail(issue, filter):
                        rlist.append(issue)

                
        return rlist

    def _pop_from_list(self, item, olist):
        """ remove one item from a list """
        nlist = []
        for oitem in olist:
            if item!=oitem:
                nlist.append(oitem)
        return nlist
    
    def _filter_fromnameemail(self, issue, filter):
        """ do the advanced filtering.
            If filter has ignorefromname then reject this issue.
            If filter has showfromname, then ignore this issue
            if not == showfromname
        """
        # if asked to be ingored return false
        if filter.has_key('ignorefromname') and filter['ignorename'] !='' and \
           ss(issue.fromname) == filter['ignorefromname']:
            return 0
        elif filter.has_key('ignoreemail') and filter['ignoreemail'] !='' and \
           ss(issue.email) == filter['ignoreemail']:
            return 0
        elif filter.has_key('showfromname') and filter['showfromname'] !='' and \
           ss(issue.fromname) != filter['showfromname']:
            return 0
        elif filter.has_key('showemail') and filter['showemail'] !='' and \
           ss(issue.email) != filter['showemail']:
            return 0
        else:
            # not trapped in anything
            return 1
    

    def doDispatchOnSubmit(self):
        """ Check if we shall dispatch emails out """
        if hasattr(self, 'dispatch_on_submit'):
            return self.dispatch_on_submit
        else:
            # default is True
            return 1

 
    def HighlightQ(self, text, q=None, highlight_html=None):
        """ Highlight a piece of a text from q """
        if highlight_html is None:
            highlight_html = '<span class="q_highlight">%s</span>'
        if q is None:
            # then look for it in REQUEST
            q = self.REQUEST.get('q',None)
        if q is None:
            return text
        else:
            transtab = string.maketrans('/ ','_ ')
            q=string.translate(q, transtab, '?&!;()<=>*#[]{}')
            mem={}
            positions = []
            last_position = 0
            n_text = ''
            for q in self.QasList(q):
                r=re.compile(q, re.IGNORECASE)
                for i in r.findall(text):
                    r= "__%s__"%self._getrandstr()
                    mem[r]=i
                    text = text.replace(i, r)
            for k,v in mem.items():
                text = text.replace(k,highlight_html%v)
            return text
    
    def _getrandstr(self,l=5):
        """ """
        pool="0123456789"
        s=''
        for i in range(l):
            s='%s%s'%(s,random.choice(list(pool)))
        return s


    def colorizeThreadChange(self, title):
        """ Make "Changed status from Open to...
            to "Changed status from <span style="color:red;">Open</span> to...
        """
        highlight_html = '<span style="color:#CC0000">%s</span>'
        for status in self.statuses:
            title = self.HighlightQ(title, status, highlight_html)
        return title

    
    def QasList(self, q):
        """ q is a string that might contain 'and' and/or 'or'.
        Remove that and make it a list. """
        r=re.compile(r"\band\b|\bor\b", re.IGNORECASE)
        return r.sub("", q).split()

    def HeadingLinks(self, display, sortname, default=0, startreversed=0):
        """Returns a hyperlink that can
           be used for resorting the listing
        """
        request = self.REQUEST
        querystring = request.QUERY_STRING

        ascending = self.www['ascarrow.gif'].tag(hspace=2)
        descending = self.www['descarrow.gif'].tag(hspace=2)
        invisible = self.www['invisiblearrow.gif'].tag(hspace=2)
        
        
        if request.has_key('reverse') and (request['reverse'] == 1 or request['reverse'] == '1'):
            reverse = 1
        elif startreversed and not (request.has_key('sortorder') and request['sortorder']==sortname):
            reverse = 1
        else:
            reverse = 0
        if request.has_key('sortorder') and request['sortorder'] == sortname or (default and not request.has_key('sortorder')):
            # have sorted by this, just let them reverse
            if reverse:
                ps = {'sortorder':sortname, 'reverse':None}
                url = self.aurl(request.URL, ps)
                url = self._addQuerystring(url)
                return """%s<a class="headinglink_current" href="%s"
                title="%s %s">%s</a>"""\
                       % (descending, url, SORT_BY, display, display )
            else:
                ps = {'sortorder':sortname, 'reverse':1}
                url = self.aurl(request.URL, ps)
                url = self._addQuerystring(url)
                return """%s<a class="headinglink_current" href="%s"
                title="%s">%s</a>"""\
                       % (ascending, url, SORT_REVERSE, display )
        else:
            if startreversed:
                ps = {'sortorder':sortname, 'reverse':1}
                url = self.aurl(request.URL, ps)
                url = self._addQuerystring(url)
                return """%s<a class="headinglink" href="%s"
                title="%s %s">%s</a>""" \
                   % (invisible, url, SORT_BY, display, display )
            else:
                ps = {'sortorder':sortname, 'reverse':None}
                url = self.aurl(request.URL, ps)
                url = self._addQuerystring(url)
                return """%s<a class="headinglink" href="%s"
                title="%s %s">%s</a>""" \
                   % (invisible, url, SORT_BY, display, display )

    def _addQuerystring(self, url):
        """ Add REQUEST querystring """
        querystring = self.REQUEST.get('QUERY_STRING','')
        if querystring is not None and querystring.strip()!='':
            url = "%s?%s"%(url, querystring)
        return url
    

    # Recent history SearchTerm
    #

    def RememberSearchTerm(self, q, yield):
        """ Stick this in a session variable """
        request = self.REQUEST
        session = request.SESSION
        key = RECENTHISTORY_SEARCHKEY
        key = self.defineInstanceSessionKey(key)
        searches = session.get(key, [])
        as_dict = {'q':q,'yield':yield}
        request.set('NotYetRecent',as_dict)
        if as_dict not in searches:
            searches.append(as_dict)
            session.set(key, searches)


    def hasRecentSearchTerms(self):
        """ check if any exists """
        request = self.REQUEST
        session = request.SESSION
        key = RECENTHISTORY_SEARCHKEY
        key = self.defineInstanceSessionKey(key)
        return session.get(key, {})!={}

    
    def getRecentSearchTerms(self):
        """ Return if any exists """
        request = self.REQUEST
        session = request.SESSION
        key = RECENTHISTORY_SEARCHKEY
        key = self.defineInstanceSessionKey(key)
        searches = session.get(key, {})
        searches.reverse()
        return searches

        #issues = []
        #for issueid in self.filterTooRecent(issueids):
        #    issues.append(getattr(self, issueid))
        #issues.reverse()
        #return issues

    # Recent history IssueVisit
    #


    def RememberIssueVisit(self, issueid):
        """ Remember that this issue has been visited """
        request = self.REQUEST
        session = request.SESSION
        key = RECENTHISTORY_ISSUEIDVISITKEY
        key = self.defineInstanceSessionKey(key)
        
        if type(issueid)!=StringType:
            # we only want objects id
            issueid = issueid.getId()

        visits = session.get(key, [])
        added_issueids = self.getRecentAddedIssues(ids=1)
        if issueid not in visits and issueid not in added_issueids:
            visits.append(issueid)
            session.set(key, visits)
            request.set('NotYetRecent',issueid)


    def hasRecentIssueVisits(self):
        """ check if any exists """
        if self.getRecentIssueVisits():
            return true
        else:
            return false
            
    def getRecentIssueVisits(self):
        """ Return if any exists """
        request = self.REQUEST
        session = request.SESSION
        key = RECENTHISTORY_ISSUEIDVISITKEY
        key = self.defineInstanceSessionKey(key)
        issueids = session.get(key, [])
        # make them objects
        issues=[]
        for issueid in self.filterTooRecent(issueids):
            try:
                issues.append(getattr(self, issueid))
            except:
                # Could have been deleted
                pass
        issues.reverse()
        return issues


    # Recent history AddedIssue
    #


    def RememberAddedIssue(self, issueid):
        """ Stick this in a session variable """
        request = self.REQUEST
        session = request.SESSION
        key = RECENTHISTORY_ADDISSUEIDKEY
        key = self.defineInstanceSessionKey(key)
        if type(issueid)!=StringType:
            # we only want objects id
            issueid = issueid.getId()
        added = session.get(key, [])
        if issueid not in added:
            added.append(issueid)
            session.set(key, added)
            request.set('NotYetRecent',issueid)

    def hasRecentAddedIssues(self):
        """ check if any exists """
        if self.getRecentAddedIssues():
            return true
        else:
            return false
            
    def getRecentAddedIssues(self, ids=0):
        """ Return if any exists """
        request = self.REQUEST
        session = request.SESSION
        key = RECENTHISTORY_ADDISSUEIDKEY
        key = self.defineInstanceSessionKey(key)
        issueids = session.get(key, [])
        # make them objects
        if ids:
            return issueids
        issues=[]
        for issueid in self.filterTooRecent(issueids):
            try:
                issues.append(getattr(self, issueid))
            except:
                # Could have been deleted
                pass
        issues.reverse()
        return issues


    def hasRecentHistory(self):
        """ check if anything is stored """
        test1 = self.hasRecentIssueVisits()
        test2 = self.hasRecentSearchTerms()
        test3 = self.hasRecentAddedIssues()
        return test1 or test2 or test3


    def filterTooRecent(self, recenthistory):
        """ Go through list and take out something too new """
        request = self.REQUEST
        too_recent_element = request.get('NotYetRecent',-1)
        n_recenthistory = []
        for each in recenthistory:
            if each != too_recent_element:
                n_recenthistory.append(each)
        return n_recenthistory
        


    # Misc scripts
    #

    def defineInstanceSessionKey(self, key):
        """ We use the default session key, but add to it for this
        issuetracker only. """
        id = self.getRoot().getId()
        return '%s-%s'%(key, id)

    def defineInstanceCookieKey(self, key):
        """ We use the default cookie key, but add to it for this
        issuetracker only. """
        # since that method is the same
        return self.defineInstanceSessionKey(key)
    
    
    # Templates
    #
    
    PTF = PageTemplateFile

    AddIssue = PTF('zpt/AddIssue', globals())
    QuickAddIssue = PTF('zpt/QuickAddIssue', globals())
    SearchForm = PTF('zpt/SearchForm', globals())
    user_searchform = PTF('zpt/user_searchform', globals())
    submitform = PTF('zpt/submitform', globals())
    main_tabs = PTF('zpt/main_tabs', globals())
    tell_a_friend = PTF('zpt/tell_a_friend', globals())
    recent_history_view = PTF('zpt/recent_history_view', globals())
    quickpreviewform = PTF('zpt/quickpreviewform', globals())
    previewform = PTF('zpt/previewform', globals())
    latestIssues = PTF('zpt/latestIssues', globals())
    ListIssues = PTF('zpt/ListIssues', globals(), __name__='ListIssues')
    security.declareProtected('View','index_html')
    index_html = PTF('zpt/index_html', globals())
    ShowIssue = PTF('zpt/ShowIssue', globals())
    ShowIssueData = PTF('zpt/ShowIssueData', globals())
    
    
   
 
IssueTracker.__dict__['About.html'] = IssueTracker.About

InitializeClass(IssueTracker)
    

###############   ISSUETRACKERISSUE    ###################

class IssueTrackerIssue(IssueTracker):
    """ Issues class as containers """

    meta_type = ISSUE_METATYPE
    icon = '%s/issue.gif'%ICON_LOCATION
    
    _properties=({'id':'title',         'type': 'string', 'mode':'w'},
                 {'id':'issuedate',     'type': 'date',   'mode':'w'},
                 {'id':'status',        'type': 'string', 'mode':'w'},
                 {'id':'type',          'type': 'string', 'mode':'w'},
                 {'id':'urgency',       'type': 'string', 'mode':'w'},
                 {'id':'sections',      'type': 'lines',  'mode':'w'},
                 {'id':'fromname',      'type': 'string', 'mode':'w'},
                 {'id':'email',         'type': 'string', 'mode':'w'},
                 {'id':'url2issue',     'type': 'string', 'mode':'w'},
                 {'id':'confidential',  'type': 'boolean','mode':'w'},
                 {'id':'hide_me',       'type': 'boolean','mode':'w'},
                 {'id':'description',   'type': 'text',   'mode':'w'},
                 {'id':'display_format','type': 'string', 'mode':'w'},
                 )
    
    security=ClassSecurityInfo()

    manage_options = (
        {'label':'Contents', 'action':'manage_main'},
        {'label':'Properties', 'action':'manage_propertiesForm'}
        )


    def __init__(self, id, title, status, issuetype, urgency, sections,
                 fromname, email, url2issue, confidential, hide_me,
                 description, display_format, issuedate=''):
        """ init an Issue object """
        self.id = str(id)
        self.title = title.strip()
        if type(issuedate) == StringType:
            issuedate = IssueTrackerUtils.DateStr2DateDate(issuedate)
        self.issuedate = issuedate
        self.status = status
        self.type = issuetype
        self.urgency = urgency
        self.sections = sections
        self.fromname = fromname
        self.email = email
        self.url2issue = url2issue
        self.confidential = confidential
        self.hide_me = hide_me
        self.description = description
        if display_format:
            self.display_format = display_format
        else:
            self.display_format = self.default_display_format
        
    security.declareProtected('View', 'index_html')
    def index_html(self, REQUEST=None):
        """ show the issue """
        if not self.confidential or self.hasManagerRole():
            return self.ShowIssue()
        else:
            self.REQUEST.RESPONSE.redirect(self.ListIssues.absolute_url())

    def IsConfidential(self):
        """ check confidential property """
        if hasattr(self, 'confidential') and self.confidential:
            return 1
        else:
            return 0
        
    def ModifyIssue(self, action='Addfollowup'):
        """ advanced change to issue properties """
        request = self.REQUEST

        #issueobject = getattr(self.getRoot(), request.issueID)
        issueobject = self
        action = ss(action)
        oldstatus = self.status
        prefix = self.issueprefix
                
        if not request.has_key('comment_description'):
            request.set('comment_description','')

        if not request.has_key('display_format'):
            request.set('display_format',self.default_display_format)

        req_manager = 1 # require manager by default
        dummychange = 0
        addfollowup = 0
        past_tense=None
        if action == 'open':
            past_tense='open'
            addfollowup = 1
        if action == 'take':
            past_tense='taken'
            addfollowup = 1
        if action == 'put on hold':
            past_tense='on hold'
            addfollowup = 1
        if action == 'reject':
            past_tense='rejected'
            addfollowup = 1
        if action == 'complete':
            past_tense='completed'
            addfollowup = 1
        if action == 'delete':
            DeleteIssue(self, self.id)
            addfollowup = 0
            redirect_url = request.URL2
        if action == 'addfollowup':
            addfollowup = 1
            dummychange = 1
            req_manager = 0

        if req_manager and not self.hasManagerRole():
            # the chosen action requires manager role,
            # but the person is not a manager
            self.login(self,request)

        if dummychange:
            # Still should trigger a new bobobase_modification_time
            status = self.status
            
        if past_tense is not None:
            self.status = past_tense
        else:
            # anything else means that the comment_description
            # cannot be empty
            if request['comment_description'].strip()=='':
                addfollowup=0
                redirect_url = request.URL1

        # most actions may perhaps add a little comment
        if addfollowup:

            genid = issueobject.generateID(4, 'issuethread'+prefix,
                                           meta_type=ISSUETHREAD_METATYPE)
            
            if action == 'addfollowup':
                gentitle = 'Added Issue followup'
            else:
                gentitle = 'Changed status from %s to %s'%\
                           (oldstatus.capitalize(), past_tense.capitalize())

            # fix variables
            try:
                fromname = request.fromname
            except:
                fromname = ''
            try:
                email = request.email
            except:
                email = ''
            try:
                display_format = request.display_format
            except:
                display_format = self.default_display_format

                
            # remember visitor
            self.set_cookie(self.getCookiekey('name'),fromname)
            self.set_cookie(self.getCookiekey('email'),email)
            self.set_cookie(self.getCookiekey('display_format'),display_format)

            # update Thread object
            title  = gentitle
            comment = request.comment_description
            threaddate = getNowdate()

            # create an Thread object
            create_method = issueobject._createThreadObject
            followupobject = create_method(genid, title, comment,
                                           threaddate, fromname,
                                           email, display_format)

            if request.has_key('notify'):
                        
                # now, create and email-alert-queue object
                # using filtered email address
                # get who to notify        
                email_addresses = self.Others2Notify(issueobject.getId(),
                                                     do='email',emailtoskip=email)

                if len(email_addresses) > 0:

                    # create id for notification
                    notifyid = self.generateID(4, prefix+"notification",
                                               meta_type='Issue Notification')

                    title = self.title
                    change = gentitle
                    issueID = self.id
                    anchorname = len(self.objectIds(ISSUETHREAD_METATYPE))
                    emails = email_addresses
                    date = getNowdate()
                    if not hasattr(self.getRoot(), issueID):
                        raise "IssueNonexisting", "This notification has no issue"
                    else:
                        notification = IssueTrackerNotification(
                                        notifyid, title, change, issueID,
                                        comment, emails, anchorname, fromname,
                                        date)
                        self._setObject(notifyid, notification)
                        notifyobject = getattr(self, notifyid)
                      
                        # use the dispatcher to try to send
                        # this notification right now.
                        request.set('loop_over',[notifyobject])
                        # there is no big deal if the dispatcher crashes here
                        # because the notification is saved and the dispatcher
                        # can be invoked some other time manually
                        if self.doDispatchOnSubmit():
                            try:
                                getattr(self,'dispatcher.html')(self, request)
                            except:
                                LOG(self.meta_type, PROBLEM,
                                    'Email could not be sent',
                                    error=sys.exc_info())
            
            
            redirect_url = '%s#i%s'%(issueobject.absolute_url(),
                                     len(issueobject.objectIds(ISSUETHREAD_METATYPE)))
        # catalog
        issueobject.reindex_object()
        # ready ! redirect!
        request.RESPONSE.redirect(redirect_url)

    def _createThreadObject(self, id, title, comment, threaddate,
                            fromname, email, display_format):
        """ Crudely create thread object. No checking. """
        thread = IssueTrackerIssueThread(id, title, comment, threaddate,
                                         fromname, email, display_format)
        self._setObject(id, thread)

        # get that object
        threadobject = getattr(self, id)

        return threadobject

    
    def Others2Notify(self, issueID, do='print', emailtoskip=None):
        """
        Returns a list of names and emails of guys to notify on this Issue. 
        Take the hide_me and confidential factor into consideration.
        
        if do == 'print':
            display the people on screen (using hide_me and confidential factor)
        if do == 'email':
            return a pure list of all email address
        """

        if emailtoskip is None:
            if self.has_cookie(self.getCookiekey('email')):
                emailtoskip = self.get_cookie(self.getCookiekey('email'))
            else:
                emailtoskip = '@@@@@'

        fromstrings = []


        issue = getattr(self.getRoot(), issueID)

        # proceed only if Manager or open-name on the issue
        if self.hasManagerRole() or not issue.hide_me:
            # self.email is the email of the thread
            # emailtoskip comes from the param/cookie if applicable
            if not issue.email == emailtoskip:
                if do =='email':
                    fromstring = getFromstring(issue.fromname, issue.email, email_must=1, email_pure=1)
                else:
                    fromstring = getFromstring(issue.fromname, issue.email)
                # fromstring looks like this now: "Peter Bengtsson <mail@peterbe.com>"
                # or ['mail@peterbe.com'] depending on 'do'
                # ...maybe
                if fromstring and fromstring not in fromstrings:
                    fromstrings.append(fromstring)

        for thread in issue.objectValues(ISSUETHREAD_METATYPE):
            if not thread.email == emailtoskip:
                if do =='email':
                    fromstring = getFromstring(thread.fromname, thread.email, email_must=1, email_pure=1)
                else:
                    fromstring = getFromstring(thread.fromname, thread.email)
                if fromstring and fromstring not in fromstrings:
                        fromstrings.append(fromstring)
        return fromstrings
    

    security.declareProtected('View', 'issue2TXT')
    def issue2TXT(self, RESPONSE):
        """ prepare an issue for download """
        if not self.confidential or self.hasManagerRole():

            if RESPONSE is not None:
                RESPONSE.setHeader('Content-type','application/data')
                RESPONSE.setHeader('Content-Disposition','inline;filename=%s.txt' % (self.id))

            return self.issue2TXT_script(self)
        else:
            self.REQUEST.RESPONSE.redirect(self.ListIssues.absolute_url())

    security.declareProtected('Access contents information', 'issue2TXT_script')
    def issue2TXT_script(self,issue):
        """ the textual representation of an issue """

        nl="\r\n"
        TXT =""
        TXT += "Export of issue (%s) %s"%(issue.id,
                                          self.ZopeTime().strftime(self.display_date))
        TXT += "%s--------------------------------%s%s"%(nl,nl,nl)
        TXT += " %s%s%s"%(issue.title,nl,nl)
        TXT += "Status:     %s%s"%(issue.status.capitalize(),nl)
        TXT += "Issue date: %s%s"%(issue.issuedate.strftime(self.display_date),nl)
        TXT += "Type:       %s%s"%(issue.type,nl)
        TXT += "Urgency:    %s%s"%(issue.urgency,nl)
        TXT += "Sections:   %s%s"%(string.join(issue.sections,', '),nl)
        TXT += "URL:        %s%s"%(issue.url2issue,nl*2)

        if issue.hide_me:
            TXT += "Name and email hidden%s"%nl
            if context.hasManagerRole():
                TXT += "Fromname:   %s%s"%(issue.fromname,nl)
                TXT += "Email:      %s%s"%(issue.email,nl*2)
        else:
            TXT += "Fromname:   %s%s"%(issue.fromname,nl)
            TXT += "Email:      %s%s"%(issue.email,nl*2)

        TXT += nl*2
        TXT += "%s%s"%(issue.description.replace('\n',nl),nl)

        threads = issue.objectValues(ISSUETHREAD_METATYPE)
        if len(threads)>0:
            TXT += "%s------------Followups------------%s"%(nl*2,nl*2)
            for thread in threads:
                TXT += "%s %s%s"%(nl,thread.title,nl*2)
                
                TXT += "Followup date: %s%s"%(thread.threaddate.strftime(self.display_date),nl)
                TXT += "Fromname:      %s%s"%(thread.fromname,nl)
                TXT += "Email:         %s%s"%(thread.email,nl*2)

                TXT += thread.comment.replace('\n',nl)
                TXT += nl*2

        return TXT

    def getOptionButtons(self):
        res=[]

        if self.status !='open':
            res.append(['IssueAction_Open','Open'])
        if self.status !='taken':
            res.append(['IssueAction_Take','Take'])
        if self.status != 'on hold':
            res.append(['IssueAction_Onhold','Put on hold'])
        if self.status != 'rejected':
            res.append(['IssueAction_Rejected','Reject'])
        if self.status != 'completed':
            res.append(['IssueAction_Complete','Complete'])
        res.append(['IssueAction_Delete','Delete',])

        return res

    def ListThreads(self,issueID):
        """ Return the followup objects. The threads. """
        req = self.REQUEST
        rlist = []
        issueobject = getattr(self.getRoot(),issueID)
        threads = issueobject.objectValues(ISSUETHREAD_METATYPE)
        return threads
    
    security.declareProtected('Manage ZCatalog Entries', 'manage_beforeDelete')
    def manage_beforeDelete(self, REQUEST=None, RESPONSE=None):
        """ uncatalog yourself """
        self.unindex_object()

    def index_object(self):
        """A common method to allow Findables to index themselves."""
        self.Catalog.catalog_object(self, '/'.join(self.getPhysicalPath()))

    def unindex_object(self):
        """A common method to allow Findables to unindex themselves."""
        self.Catalog.uncatalog_object('/'.join(self.getPhysicalPath()))

    # Templates
    PTF = PageTemplateFile
    
    shownewIssueURL = PTF('zpt/shownewIssueURL', globals())
    manager_options = PTF('zpt/manager_options', globals())
    anonymous_options = PTF('zpt/anonymous_options', globals())
    form_followup = PTF('zpt/form_followup', globals())
    form_open = PTF('zpt/form_open', globals())
    form_take = PTF('zpt/form_take', globals())
    form_onhold = PTF('zpt/form_onhold', globals())
    form_reject = PTF('zpt/form_reject', globals())
    form_delete = PTF('zpt/form_delete', globals())
    form_complete = PTF('zpt/form_complete', globals())
    followup_preview = PTF('zpt/followup_preview', globals())


InitializeClass(IssueTrackerIssue)


############### ISSUETRACKERISSUETHREAD ##################

class IssueTrackerIssueThread(SimpleItem.SimpleItem, IssueTrackerIssue):
    """ Issuethreads class """

    meta_type = ISSUETHREAD_METATYPE
    icon = '%s/issuethread.gif'%ICON_LOCATION
    
    meta_types = []

    _properties=({'id':'title',         'type': 'string', 'mode':'w'},
                 {'id':'comment',       'type': 'text',   'mode':'w'},
                 {'id':'threaddate',    'type': 'date',   'mode':'w'},
                 {'id':'fromname',      'type': 'string', 'mode':'w'},
                 {'id':'email',         'type': 'string', 'mode':'w'},
                 {'id':'display_format','type': 'string', 'mode':'w'},
                 )
    
    security=ClassSecurityInfo()

    manage_options = (
        {'label':'Properties', 'action':'manage_propertiesForm'},
        )

    def __init__(self, id, title, comment, threaddate, fromname, email,
                 display_format=None, REQUEST=None):
        """ create thread """
        self.id = str(id)
        self.title = title
        self.comment = comment
        if type(threaddate) == StringType:
            threaddate = DateTime(threaddate)
        self.threaddate = threaddate
        self.fromname = fromname
        self.email = email
        if display_format:
            self.display_format = display_format
        else:
            self.display_format = self.default_display_format

    


InitializeClass(IssueTrackerIssueThread)




############### ISSUETRACKERNOTIFICATION ###################

class IssueTrackerNotification(SimpleItem.SimpleItem,
                               IssueTracker
                               ):
    """ Issue Tracker Notification """

    meta_type = NOTIFICATION_META_TYPE
    icon = '%s/notification.gif'%ICON_LOCATION

    security=ClassSecurityInfo()
    
    _properties=({'id':'title',       'type': 'string',  'mode':'w'},
                 {'id':'change',      'type': 'string',  'mode':'w'},
                 {'id':'issueID',     'type': 'string',  'mode':'w'},
                 {'id':'comment',     'type': 'text',    'mode':'w'},
                 {'id':'emails',      'type': 'lines',   'mode':'w'},
                 {'id':'anchorname',  'type': 'string',  'mode':'w'},
                 {'id':'fromname',    'type': 'string',  'mode':'w'},
                 {'id':'date',        'type': 'date',    'mode':'w'},
                 {'id':'dispatched',  'type': 'boolean', 'mode':'w'},
                 )

    manage_options = (
        {'label':'Properties', 'action':'manage_propertiesForm'},
        )

    def __init__(self, id, title, change, issueID, comment, emails,
                 anchorname, fromname, date=None, dispatched=0, REQUEST=None):
        """ create notification """
        issueID = str(issueID)
        self.id = str(id)
        self.title = title
        self.change = change
        self.issueID = issueID
        self.comment = comment
        self.emails = emails
        self.anchorname = anchorname
        self.fromname = fromname
        self.dispatched = dispatched
        if date is None:
            self.date = getNowdate()
        else:
            self.date = date

    def dispatch(self):
        """ dispatch self """
        self.REQUEST.set('loop_over',self)
        getattr(self, 'dispatcher.html')
        
    def MarkNotificationDispatch(self):
        """ set as dispatched """
        self.dispatched = 1
        
InitializeClass(IssueTrackerNotification)

